class MNUtil {
  static themeColor = {
    Gray: UIColor.colorWithHexString("#414141"),
    Default: UIColor.colorWithHexString("#FFFFFF"),
    Dark: UIColor.colorWithHexString("#000000"),
    Green: UIColor.colorWithHexString("#E9FBC7"),
    Sepia: UIColor.colorWithHexString("#F5EFDC")
  }
  static init(mainPath){
    this.mainPath = mainPath
    this.MNUtilVersion = this.getMNUtilVersion()
    const renderer = new marked.Renderer();
    renderer.code = (code, language) => {
      const validLang = hljs.getLanguage(language) ? language : 'plaintext';
      const uuid = NSUUID.UUID().UUIDString()
      if (validLang === 'plaintext') {
        return `<pre><div class="code-header" contenteditable="false"><span>${validLang}</span><button onclick="copyToClipboard('${uuid}')">复制</button></div><code class="hljs ${validLang}"  id="${uuid}">${code}</code></pre>`;
      }
      const highlightedCode = hljs.highlight(code, { language: validLang }).value;
      return `<pre><div class="code-header" contenteditable="false"><span>${validLang}</span><button onclick="copyToClipboard('${uuid}')">复制</button></div><code class="hljs ${validLang}" id="${uuid}">${highlightedCode}</code></pre>`;
    };
    marked.setOptions({ renderer });
  }
  static get version(){
    if (!this.mnVersion) {
      this.mnVersion = this.appVersion()
    }
    return this.mnVersion
  }
  static get app(){
    if (!this.appInstance) {
      this.appInstance = Application.sharedInstance()
    }
    return this.appInstance
  }
  static get db(){
    if (!this.data) {
      this.data = Database.sharedInstance()
    }
    return this.data
  }
  static get currentWindow(){
    //关闭mn4后再打开，得到的focusWindow会变，所以不能只在init做一遍初始化
    return this.app.focusWindow
  }
  static get studyController(){
    return this.app.studyController(this.currentWindow)
  }
  static get studyView() {
    return this.app.studyController(this.currentWindow).view
  }
  static get studyMode(){
    return this.studyController.studyMode
  }
  static get readerController() {
    return this.studyController.readerController
  }
  static get notebookController(){
    return this.studyController.notebookController
  }
  static get docControllers() {
    return this.studyController.readerController.documentControllers
  }
  static get currentDocController() {
    return this.studyController.readerController.currentDocumentController
  }
  static get mindmapView(){
    return this.studyController.notebookController.mindmapView
  }
  static get selectionText(){
    return this.currentDocController.selectionText
  }
  static get currentNotebookId() {
    return this.studyController.notebookController.notebookId
  }
  static get currentNotebook() {
    return this.getNoteBookById(this.currentNotebookId)
  }
  static get currentDocmd5(){
    try {
      const { docMd5 } = this.currentDocController
      if (docMd5 && docMd5.length === 32) return "00000000"
      else return docMd5
    } catch {
      return undefined
    }
  }
  static get isZH(){
    return NSLocale.preferredLanguages()[0].startsWith("zh")
  }

  static get currentThemeColor(){
    return this.themeColor[this.app.currentTheme]
  }
  static get clipboardText() {
    return UIPasteboard.generalPasteboard().string
  }
  static get documentFolder() {
    return this.app.documentPath
  }
  static get tempFolder() {
    return this.app.tempPath
  }
  static get splitLine() {
    let study = this.studyController
    let studyFrame = study.view.bounds
    let readerFrame = study.readerController.view.frame
    let hidden = study.readerController.view.hidden//true代表脑图全屏
    let rightMode = study.rightMapMode
    let fullWindow = readerFrame.width == studyFrame.width
    if (hidden || fullWindow) {
      return undefined
    }
    if (rightMode) {
      let splitLine = readerFrame.x+readerFrame.width
      return splitLine
    }else{
      let splitLine = readerFrame.x
      return splitLine
    }
  }
  static appVersion() {
    let info = {}
    let version = parseFloat(this.app.appVersion)
    if (version >= 4) {
      info.version = "marginnote4"
    }else{
      info.version = "marginnote3"
    }
    switch (this.app.osType) {
      case 0:
        info.type = "iPadOS"
        break;
      case 1:
        info.type = "iPhoneOS"
        break;
      case 2:
        info.type = "macOS"
      default:
        break;
    }
    return info
  }
  static getMNUtilVersion(){
    let res = this.readJSON(this.mainPath+"/mnaddon.json")
    return res.version
    // this.copyJSON(res)
  }
  static showHUD(message,duration=2) {
    this.app.showHUD(message,this.currentWindow,2)
  }
  static async confirm(mainTitle,subTitle){
    return new Promise((resolve, reject) => {
      UIAlertView.showWithTitleMessageStyleCancelButtonTitleOtherButtonTitlesTapBlock(
        mainTitle,subTitle,0,"Cancel",["Confirm"],
        (alert, buttonIndex) => {
          // MNUtil.copyJSON({alert:alert,buttonIndex:buttonIndex})
          resolve(buttonIndex)
        }
      )
    })
  }
  static copy(text) {
    UIPasteboard.generalPasteboard().string = text
  }
  static copyJSON(object) {
    UIPasteboard.generalPasteboard().string = JSON.stringify(object,null,2)
  }
  /**
   * 
   * @param {NSData} imageData 
   */
  static copyImage(imageData) {
    UIPasteboard.generalPasteboard().setDataForPasteboardType(imageData,"public.png")
  }
  static openURL(url){
    this.app.openURL(NSURL.URLWithString(url));
  }
  static getNoteById(noteid) {
    let note = this.db.getNoteById(noteid)
    if (note) {
      return note
    }else{
      this.showHUD("Note not exist!")
      return undefined
    }
  }
  static getNoteBookById(notebookId) {
    let notebook = this.db.getNotebookById(notebookId)
    return notebook
  }
  static getDocById(md5) {
    let doc = this.db.getDocumentById(md5)
    return doc
  }
  /**
   * 
   * @param {String} url 
   * @returns {String}
   */
  static getNoteIdByURL(url) {
    let targetNoteId = url.trim()
    if (/^marginnote\dapp:\/\/note\//.test(targetNoteId)) {
      targetNoteId = targetNoteId.slice(22)
    }
    return targetNoteId
  }
  static isfileExists(path) {
    return NSFileManager.defaultManager().fileExistsAtPath(path)
  }
  static genFrame(x,y,width,height) {
    return {x:x,y:y,width:width,height:height}
  }
  static setFrame(view,x,y,width,height){
    view.frame = {x:x,y:y,width:width,height:height}
  }
  static parseWinRect(winRect) {
    let rectArr = winRect.replace(/{/g, '').replace(/}/g, '').replace(/\s/g, '').split(',')
    let X = Number(rectArr[0])
    let Y = Number(rectArr[1])
    let H = Number(rectArr[3])
    let W = Number(rectArr[2])
    return this.genFrame(X, Y, W, H)
  }
  static async animate(func,time = 0.2) {
    return new Promise((resolve, reject) => {
      UIView.animateWithDurationAnimationsCompletion(time,func,()=>(resolve()))
    })
  }
  static checkSender(sender,window = this.currentWindow){
    return this.app.checkNotifySenderInWindow(sender, window)
  }
  static async delay (seconds) {
    return new Promise((resolve, reject) => {
      NSTimer.scheduledTimerWithTimeInterval(seconds, false, function () {
        resolve()
      })
    })
  }
  /**
   * 
   * @param {UIView} view 
   */
  static isDescendantOfStudyView(view){
    return view.isDescendantOfView(this.studyView)
  }
  static addObserver(observer,selector,name){
    NSNotificationCenter.defaultCenter().addObserverSelectorName(observer, selector, name);
  }
  static removeObserver(observer,name){
    NSNotificationCenter.defaultCenter().removeObserverName(observer, name);
  }
  static removeObservers(object,notifications) {
    notifications.forEach(notification=>{
      NSNotificationCenter.defaultCenter().removeObserverName(object, notification);
    })
  }
  static refreshAddonCommands(ignoreImport = false){
    this.studyController.refreshAddonCommands()
  }
  static focusNoteInMindMapById(noteId){
    let note = this.getNoteById(noteId)
    if (note.notebookId !== this.currentNotebookId) {
      this.showHUD("Note not in current notebook")
      return
    }
    this.studyController.focusNoteInMindMapById(noteId)
  }
  static focusNoteInFloatMindMapById(noteId){
    this.studyController.focusNoteInFloatMindMapById(noteId)
  }
  static focusNoteInDocumentById(noteId){
    this.studyController.focusNoteInDocumentById(noteId)
  }
  static isValidJSON(jsonString){
    // return NSJSONSerialization.isValidJSONObject(result)
     try{
         var json = JSON.parse(jsonString);
         if (json && typeof json === "object") {
             return true;
         }
     }
     catch(e){
         return false;
     }
     return false;
  }
  static getValidJSON(text) {
    try {
      let original = text
      text = text.replace("}{",",")
    if (text.endsWith(':')) {
      text = text+'""}'
    }
    if (this.isValidJSON(text)) {
      return JSON.parse(text)
    }else if (this.isValidJSON(text+"\"}")){
      return JSON.parse(text+"\"}")
    }else if (this.isValidJSON(text+"}")){
      return JSON.parse(text+"}")
    }else if (!text.startsWith('{') && text.endsWith('}')){
      return JSON.parse("{"+text)
    }else{
      // this.copy(original)
      return undefined
    }
    } catch (error) {
      this.showHUD("Error in getValidJSON: "+error)
      this.copy(text)
      return undefined
    }
  }
  static mergeWhitespace(str) {
      // 先将多个连续的换行符替换为单个换行符
      var tempStr = str.replace(/\n+/g, '\n');
      // 再将其它的空白符（除了换行符）替换为单个空格
      return tempStr.replace(/[\r\t\f\v ]+/g, ' ').trim();
  }
  static undoGrouping(f,notebookId = this.currentNotebookId){
    UndoManager.sharedInstance().undoGrouping(
      String(Date.now()),
      notebookId,
      f
    )
    this.app.refreshAfterDBChanged(notebookId)
  }
  static getImage(path,scale=2) {
    return UIImage.imageWithDataScale(NSData.dataWithContentsOfFile(path), scale)
  }
  static getFile(path) {
    return NSData.dataWithContentsOfFile(path)
  }
  /**
   * 
   * @param {string} fullPath 
   * @returns {string}
   */
  static getFileName(fullPath) {
      // 找到最后一个'/'的位置
      let lastSlashIndex = fullPath.lastIndexOf('/');

      // 从最后一个'/'之后截取字符串，得到文件名
      let fileName = fullPath.substring(lastSlashIndex + 1);

      return fileName;
  }
  static getMediaByHash(hash) {
    let image = this.db.getMediaByHash(hash)
    return image
  }
  static getPopoverAndPresent(sender,commandTable,width=100) {
    var menuController = MenuController.new();
    menuController.commandTable = commandTable
    menuController.rowHeight = 35;
    menuController.preferredContentSize = {
      width: width,
      height: menuController.rowHeight * menuController.commandTable.length
    };
    var popoverController = new UIPopoverController(menuController);
    var r = sender.convertRectToView(sender.bounds,this.studyView);
    popoverController.presentPopoverFromRect(r, this.studyView, 1 << 1, true);
    return popoverController
  }
  static postNotification(name,userInfo) {
    NSNotificationCenter.defaultCenter().postNotificationNameObjectUserInfo(name, this.currentWindow, userInfo)
  }
  static hexColorAlpha(hex,alpha) {
    let color = UIColor.colorWithHexString(hex)
    return alpha!==undefined?color.colorWithAlphaComponent(alpha):color
  }
  static genNSURL(url) {
    return NSURL.URLWithString(url)
  }
  static data2string(data){
    let test = CryptoJS.enc.Base64.parse(data.base64Encoding())
    let textString = CryptoJS.enc.Utf8.stringify(test);
    return textString
  }
  static MD5(str){
    let md5 = CryptoJS.MD5(str).toString();
    return md5
  }
  static md2html(md){
    return marked.parse(md)
  }
  static escapeString(str) {
    return str.replace(/\\/g, '\\\\') // Escape backslashes
              .replace(/\`/g, '\\`') // Escape backticks
              .replace(/\$\{/g, '\\${') // Escape template literal placeholders
              .replace(/\r/g, '\\r') // Escape carriage returns
              .replace(/\n/g, '\\n') // Escape newlines
              .replace(/'/g, "\\'")   // Escape single quotes
              .replace(/"/g, '\\"');  // Escape double quotes
  }
  static readJSON(path){
    let data = NSData.dataWithContentsOfFile(path)
    const res = NSJSONSerialization.JSONObjectWithDataOptions(
      data,
      1<<0
    )
    return res
  }
  static writeJSON(path,data){
    NSData.dataWithStringEncoding(
      JSON.stringify(data, undefined, 2),
      4
    ).writeToFileAtomically(path, false)
  }
  static saveFile(file, UTI) {
    this.app.saveFileWithUti(file, UTI)
  }
  /**
   * 
   * @param {string[]} UTI 
   * @returns 
   */
  static async importFile(UTI){
    return new Promise((resolve, reject) => {
      this.app.openFileWithUTIs(UTI,this.studyController,(path)=>{
        resolve(path)
      })
    })
  }
  /**
   * 
   * @param {T[]} arr 
   * @param {boolean} noEmpty 
   * @returns {T[]}
   */
 static unique (arr, noEmpty = false){
  let ret = []
  if (arr.length <= 1) ret = arr
  else ret = Array.from(new Set(arr))
  if (noEmpty) ret = ret.filter(k => k)
  return ret
}
  static typeOf(note){
    if (typeof note === "string") {
      if (/^marginnote\dapp:\/\/note\//.test(note.trim())) {
        return "NoteURL"
      }
      return "string"
    }
    if (note instanceof MNNote) {
      return "MNNote"
    }
    if (note.noteId) {
      return "MbBookNote"
    }
    if ("title" in note || "content" in note) {
      return "NoteConfig"
    }
    return typeof note

  }
}

class MNButton{
  static get highlightColor(){
    return UIColor.blendedColor(
      UIColor.colorWithHexString("#2c4d81").colorWithAlphaComponent(0.8),
      MNUtil.app.defaultTextColor,
      0.8
    );
  }
  static new(config = {}){
    let newButton = UIButton.buttonWithType(0);
    newButton.autoresizingMask = (1 << 0 | 1 << 3);
    newButton.layer.masksToBounds = true;
    newButton.setTitleColorForState(UIColor.whiteColor(),0);
    newButton.setTitleColorForState(this.highlightColor, 1);
    let radius = ("radius" in config) ? config.radius : 8
    newButton.layer.cornerRadius = radius;
    this.setConfig(newButton, config)
    return newButton
  }
//   static createButton(superview,config) {
//     let button = UIButton.buttonWithType(0);
//     button.autoresizingMask = (1 << 0 | 1 << 3);
//     button.setTitleColorForState(UIColor.whiteColor(),0);
//     button.setTitleColorForState(this.highlightColor, 1);
//     button.backgroundColor = this.hexColorAlpha("#9bb2d6",0.8)
//     button.layer.cornerRadius = 8;
//     button.layer.masksToBounds = true;
//     button.titleLabel.font = UIFont.systemFontOfSize(16);
//     if (superview) {
//       superview.addSubview(button)
//     }
//     this.setConfig(button, config)
//     return button
// }
  static hexColorAlpha(hex,alpha) {
    let color = UIColor.colorWithHexString(hex)
    return alpha!==undefined?color.colorWithAlphaComponent(alpha):color
  }  
  static setColor(button,hexColor,alpha = 1.0){
    button.backgroundColor = this.hexColorAlpha(hexColor, alpha)
  }
  static setTitle(button,title,font = 16,bold= false){
    button.setTitleForState(title,0)
    if (bold) {
      button.titleLabel.font = UIFont.boldSystemFontOfSize(font)
    }else{
      button.titleLabel.font = UIFont.systemFontOfSize(font)
    }
  }
  static setImage(button,path){
    button.setImageForState(MNUtil.getImage(path),0)
  }
  static setOpacity(button,opacity){
    button.layer.opacity = opacity
  }
  static setRadius(button,radius = 8){
    button.layer.cornerRadius = radius;
  }
  static setConfig(button,config){
    if ("color" in config) {
      this.setColor(button, config.color, config.alpha)
    }
    if ("title" in config) {
      this.setTitle(button, config.title, config.font, config.bold)
    }
    if ("opacity" in config) {
      this.setOpacity(button, config.opacity)
    }
    if ("radius" in config) {
      this.setRadius(button,config.radius)
    }
  }
}

class MNNote{
  /** @type {MbBookNote} */
  note
  /**
   * 
   * @param {MbBookNote|string|object} note 
   */
  constructor(note) {
    switch (MNUtil.typeOf(note)) {
      case 'MbBookNote':
        this.note = note
        break;
      case 'NoteURL':
        let NoteFromURL = MNUtil.getNoteBookById(MNUtil.getNoteIdByURL(note))
        if (NoteFromURL) {
          this.note = NoteFromURL
        }
      case 'string':
        let targetNoteId = note.trim()
        let targetNote = MNUtil.getNoteById(targetNoteId)
        if (targetNote) {
          this.note = targetNote
        }
        // MNUtil.showHUD(this.note.noteId)
        break;
        // MNUtil.copyJSON(targetNoteId+":"+this.note.noteId)
      case "NoteConfig":
        let config = note
        let notebook = MNUtil.currentNotebook
        let title = config.title ?? ""
        // MNUtil.showHUD("new note")
        // MNUtil.copyJSON(note)
        this.note = Note.createWithTitleNotebookDocument(title, notebook, MNUtil.currentDocController.document)
        if (config.content) {
          if (config.markdown) {
            this.note.appendMarkdownComment(config.content)
          }else{
            this.note.appendTextComment(config.content)
          }
        }
        if (config.color !== undefined) {
          this.note.colorIndex = config.color
        }
        break;
      default:
        break;
    }
  }
  /**
   * 
   * @param {MbBookNote|string|object} note 
   */
  static new(note){
    switch (MNUtil.typeOf(note)) {
      case 'MbBookNote':
        return new MNNote(note)
      case 'NoteURL':
        let NoteFromURL = MNUtil.getNoteBookById(MNUtil.getNoteIdByURL(note))
        if (NoteFromURL) {
          return new MNNote(NoteFromURL)
        }
        return undefined
      case 'string':
        let targetNoteId = note.trim()
        let targetNote = MNUtil.getNoteById(targetNoteId)
        if (targetNote) {
          return new MNNote(targetNote)
        }
        return undefined
      case "NoteConfig":
        let config = note
        let notebook = MNUtil.currentNotebook
        let title = config.title ?? ""
        // MNUtil.showHUD("new note")
        // MNUtil.copyJSON(note)
        let newNote = Note.createWithTitleNotebookDocument(title, notebook, MNUtil.currentDocController.document)
        if (config.content) {
          if (config.markdown) {
            newNote.appendMarkdownComment(config.content)
          }else{
            newNote.appendTextComment(config.content)
          }
        }
        if (config.color !== undefined) {
          newNote.colorIndex = config.color
        }
        return new MNNote(newNote)
      default:
        return undefined
    }
  }
  get noteId() {
    return this.note.noteId
  }
  get notebookId() {
    return this.note.notebookId
  }
  get childNotes() {
    return this.note.childNotes?.map(k => new MNNote(k)) ?? []
  }
  get parentNote() {
    return this.note.parentNote && new MNNote(this.note.parentNote)
  }
  get noteURL(){
    return MNUtil.version.version+'app://note/'+this.note.noteId
  }
  /**
   * 
   * @returns {{descendant:MNNote[],treeIndex:number[][]}}
   */
  get descendantNodes() {
    const { childNotes } = this
    if (!childNotes.length) {
      return {
        descendant: [],
        treeIndex: []
      }
    } else {
      /**
       * 
       * @param {MNNote[]} nodes 
       * @param {number} level 
       * @param {number[]} lastIndex 
       * @param {{descendant:MNNote[],treeIndex:number[][]}} ret 
       * @returns 
       */
      function down(
        nodes,
        level = 0,
        lastIndex = [],
        ret = {
          descendant: [],
          treeIndex: []
        }
      ) {
        level++
        nodes.forEach((node, index) => {
          ret.descendant.push(node)
          lastIndex = lastIndex.slice(0, level - 1)
          lastIndex.push(index)
          ret.treeIndex.push(lastIndex)
          if (node.childNotes?.length) {
            down(node.childNotes, level, lastIndex, ret)
          }
        })
        return ret
      }
      return down(childNotes)
    }
  }
  get ancestorNodes() {
    /**
     * 
     * @param {MNNote} node 
     * @param {MNNote[]} ancestorNodes 
     * @returns 
     */
    function up(node, ancestorNodes) {
      if (node.note.parentNote) {
        const parentNode = new MNNote(node.note.parentNote)
        ancestorNodes = up(parentNode, [...ancestorNodes, parentNode])
      }
      return ancestorNodes
    }
    return up(this, [])
  }
  get notes() {
    return this.note.comments.reduce(
      (acc, cur) => {
        cur.type == "LinkNote" && acc.push(MNUtil.db.getNoteById(cur.noteid))
        return acc
      },
      [this.note]
    )
  }
  get titles() {
    return MNUtil.unique(this.note.noteTitle?.split(/\s*[;；]\s*/) ?? [], true)
  }
  /**
   * 
   * @param {string[]} titles 
   * @returns 
   */
  set titles(titles) {
    const newTitle = MNUtil.unique(titles, true).join("; ")
    if (this.note.excerptText === this.note.noteTitle) {
      this.note.noteTitle = newTitle
      this.note.excerptText = newTitle
    } else {
      this.note.noteTitle = newTitle
    }
  }
  get isOCR() {
    if (this.note.excerptPic?.paint) {
      return this.note.textFirst
    }
  }
  get title() {
    return this.note.noteTitle ?? ""
  }
  get noteTitle() {
    return this.note.noteTitle ?? ""
  }
  /**
   * 
   * @param {string} title
   * @returns 
   */
  set title(title) {
    this.note.noteTitle = title
  }
  /**
   * 
   * @param {string} title
   * @returns 
   */
  set noteTitle(title) {
    this.note.noteTitle = title
  }
  get excerptText(){
    return this.note.excerptText
  }
  set excerptText(text){
    this.note.excerptText = text
  }
  get mainExcerptText() {
    return this.note.excerptText ?? ""
  }
  /**
   * 
   * @param {string} text
   * @returns 
   */
  set mainExcerptText(text) {
    this.note.excerptText = text
  }
  get excerptPic(){
    return this.note.excerptPic
  }
  get excerptPicData(){
    let imageData = MNUtil.getMediaByHash(this.note.excerptPic.paint)
    return imageData
  }
  get colorIndex(){
    return this.note.colorIndex
  }
  set colorIndex(index){
    this.note.colorIndex = index
  }
  get fillIndex(){
    return this.note.fillIndex
  }
  set fillIndex(index){
    this.note.fillIndex = index
  }
  get createDate(){
    return this.note.createDate
  }
  get modifiedDate(){
    return this.note.modifiedDate
  }
  get linkedNotes(){
    return this.note.linkedNotes
  }
  get summaryLinks(){
    return this.summaryLinks
  }
  get textFirst(){
    return this.textFirst
  }
  /**
   * get all tags, without `#`
   * @returns {string[]}
   */
  get tags() {
    try {
    // MNUtil.showHUD("length: "+this.note.comments.length)
    const tags = this.note.comments.reduce((acc, cur) => {
      // MNUtil.showHUD("cur.type")
      if (cur.type == "TextNote" && cur.text.startsWith("#")) {
        acc.push(...cur.text.split(/\s+/).filter(k => k.startsWith("#")))
      }
      return acc
    }, [])
    return tags.map(k => k.slice(1))
      
    } catch (error) {
      MNUtil.showHUD(error)
      return []
    }
  }
  /**
   * 
   * @returns {NoteComment[]}
   */
  get comments(){
    return this.note.comments
  }
  /**
   * set tags, will remove all old tags
   * 
   * @param {string[]} tags
   * @returns 
   */
  set tags(tags) {
    this.tidyupTags()
    tags = MNUtil.unique(tags, true)
    const lastComment = this.note.comments[this.note.comments.length - 1]
    if (lastComment?.type == "TextNote" && lastComment.text.startsWith("#")) {
      this.note.removeCommentByIndex(this.note.comments.length - 1)
    }
    this.appendTextComments(tags.map(k => `#${k}`).join(" "))
  }
  /**
   * @returns {{ocr:string[],html:string[],md:string[]}}
   */
  get excerptsTextPic() {
    return this.notes.reduce(
      (acc, cur) => {
        Object.entries(MNNote.getNoteExcerptTextPic(cur)).forEach(([k, v]) => {
          if (k in acc) acc[k].push(...v)
        })
        return acc
      },
      {
        ocr: [],
        html: [],
        md: []
      }
    )
  }
  /**
   * @returns {{html:string[],md:string[]}}
   */
  get commentsTextPic() {
    return this.note.comments.reduce(
      (acc, cur) => {
        if (cur.type === "PaintNote") {
          const imgs = MNNote.exportPic(cur)
          if (imgs)
            Object.entries(imgs).forEach(([k, v]) => {
              if (k in acc) acc[k].push(v)
            })
        } else if (cur.type == "TextNote" || cur.type == "HtmlNote") {
          const text = cur.text.trim()
          if (text && !text.includes("marginnote3app") && !text.startsWith("#"))
            Object.values(acc).map(k => k.push(text))
        }
        return acc
      },
      {
        html: [],
        md: []
      }
    )
  }
  /** @returns {string[]} */
 get excerptsText() {
    return this.notes.reduce((acc, note) => {
      const text = note.excerptText?.trim()
      if (text) {
        if (!note.excerptPic?.paint || this.isOCR) {
          acc.push(text)
        }
      }
      return acc
    }, [])
  }
  /**
   * get all comment text
   * @returns {string[]}
   */
  get commentsText() {
    return this.note.comments.reduce((acc, cur) => {
      if (cur.type == "TextNote" || cur.type == "HtmlNote") {
        const text = cur.text.trim()
        if (text && !text.includes("marginnote3app") && !text.startsWith("#"))
          acc.push(text)
      }
      return acc
    }, [])
  }
  /**
   * get all text and pic note will be OCR or be transformed to base64
   */
  get allTextPic() {
    const retVal = MNNote.getNoteExcerptTextPic(this.note)
    this.note.comments.forEach(k => {
      if (k.type === "PaintNote") {
        const imgs = MNNote.exportPic(k)
        if (imgs)
          Object.entries(imgs).forEach(([k, v]) => {
            if (k in retVal) retVal[k].push(v)
          })
      } else if (k.type == "TextNote" || k.type == "HtmlNote") {
        const text = k.text.trim()
        if (text) Object.values(retVal).map(k => k.push(text))
      } else if (k.type == "LinkNote") {
        const note = MNUtil.db.getNoteById(k.noteid)
        if (note)
          Object.entries(MNNote.getNoteExcerptTextPic(note)).forEach(([k, v]) => {
            if (k in retVal) retVal[k].push(...v)
          })
      }
    })
    return {
      html: retVal.html.join("\n\n"),
      ocr: retVal.ocr.join("\n\n"),
      md: retVal.md.join("\n\n")
    }
  }
  /**
   * Get all text
   */
  get allText() {

    const { mainExcerptText } = this
    const retVal =
      mainExcerptText && (!this.note.excerptPic?.paint || this.isOCR)
        ? [mainExcerptText]
        : []
    this.note.comments.forEach(k => {
      if (k.type == "TextNote" || k.type == "HtmlNote") {
        const text = k.text.trim()
        if (text) retVal.push(text)
      } else if (k.type == "LinkNote") {
        const note = MNUtil.db.getNoteById(k.noteid)
        const text = note?.excerptText?.trim()
        if (text && (!note?.excerptPic?.paint || this.isOCR)) retVal.push(text)
      }
    })
    return retVal.join("\n\n")
  }
  /**
   * Get all text.
   */
  get excerptsCommentsText() {
    const { mainExcerptText } = this
    const retVal =
      mainExcerptText && (!this.note.excerptPic?.paint || this.isOCR)
        ? [mainExcerptText]
        : []
    this.note.comments.forEach(k => {
      if (k.type == "TextNote" || k.type == "HtmlNote") {
        const text = k.text.trim()
        if (text && !text.includes("marginnote3app") && !text.startsWith("#"))
          retVal.push(text)
      } else if (k.type == "LinkNote") {
        const note = MNUtil.db.getNoteById(k.noteid)
        const text = note?.excerptText?.trim()
        if (text && (!note?.excerptPic?.paint || this.isOCR)) retVal.push(text)
      }
    })
    return retVal
  }
  allNoteText(){
    return this.note.allNoteText()
  }
  paste(){
    this.note.paste()
  }
  /**
   * 
   * @param {MbBookNote|MNNote|string} note 
   */
  merge(note){
    switch (MNUtil.typeOf(note)) {
      case "MbBookNote":
        this.note.merge(note)
        break;
      case "MNNote":
        this.note.merge(note.note)
        break;
      case "NoteURL":
        let noteFromURL = MNUtil.getNoteById(MNUtil.getNoteIdByURL(note))
        if (noteFromURL) {
          this.note.merge(noteFromURL)
        }else{
          MNUtil.showHUD("Note not exist!")
        }
      case "string":
        let targetNote = MNUtil.getNoteById(note)
        if (targetNote) {
          this.note.merge(targetNote)
        }else{
          MNUtil.showHUD("Note not exist!")
        }
        break
      default:
        break;
    }
  }
  /**
   * 
   * @param {MbBookNote|MNNote|string} note 
   */
  addChild(note){
    switch (MNUtil.typeOf(note)) {
      case "NoteURL":
        let noteFromURL = MNUtil.getNoteById(MNUtil.getNoteIdByURL(note))
        if (noteFromURL) {
          this.note.addChild(noteFromURL)
        }else{
          MNUtil.showHUD("Note not exist!")
        }
        break;
      case "string":
        let targetNote = MNUtil.getNoteById(note)
        if (targetNote) {
          this.note.addChild(targetNote)
        }else{
          MNUtil.showHUD("Note not exist!")
        }
        break;
      case "MNNote":
        this.note.addChild(note.note)
        break;
      case "MbBookNote":
        this.note.addChild(note)
        break;
      default:
        break;
    }
  }
  /**
   * 
   * @param {MbBookNote|MNNote} targetNote 
   */
  addAsChildNote(targetNote,colorInheritance=false) {
    MNUtil.undoGrouping(()=>{
      if (colorInheritance) {
        targetNote.colorIndex = this.note.colorIndex
      }
      this.addChild(targetNote)
    })
  }
  /**
   * 
   * @param {MbBookNote|MNNote} targetNote 
   */
  addAsBrotherNote(targetNote,colorInheritance=false) {
    if (!this.note.parentNote) {
      MNUtil.showHUD("No parent note!")
      return
    }
    let parent = this.parentNote
    MNUtil.undoGrouping(()=>{
      if (colorInheritance) {
        targetNote.colorIndex = this.note.colorIndex
      }
      parent.addChild(targetNote)
    })
  }
  /**
   * 
   * @param {{title:String,content:String,markdown:Boolean,color:Number}} config 
   * @returns {MNNote}
   */
  createChildNote(config) {
    let child
    MNUtil.undoGrouping(()=>{
      try {
      child = MNNote.new(config)
      // this.focusInMindMap()
      // this.note.addChild(child.note)
      this.addChild(child)
        
      } catch (error) {
        MNUtil.showHUD(error)
      }
    })
    return child
  }
  /**
   * 
   * @param {{title:String,content:String,markdown:Boolean,color:Number}} config 
   * @returns {MNNote}
   */
  createBrotherNote(config) {
    if (!this.parentNote) {
      MNUtil.showHUD("No parent note!")
      return
    }
    let child
    let parent = this.parentNote
    MNUtil.undoGrouping(()=>{
      child = MNNote.new(config)
      parent.addChild(child)
    })
    return child
  }
  /**
   * Append text comments as much as you want.
   * @param {string[]} comments
   * @example
   * node.appendTextComments("a", "b", "c")
   */
  appendTextComments(...comments) {
    comments = MNUtil.unique(comments, true)
    const existComments = this.note.comments.filter(k => k.type === "TextNote")
    comments.forEach(comment => {
      if (
        comment &&
        existComments.every(k => k.type === "TextNote" && k.text !== comment)
      ) {
        this.note.appendTextComment(comment)
      }
    })
    return this
  }
  /**
   * 
   * @param  {string} comment
   * @returns 
   */
  appendMarkdownComment(comment){
    this.note.appendMarkdownComment(comment)
  }
  /**
   * 
   * @param  {string} comment
   * @returns 
   */
  appendTextComment(comment){
    this.note.appendTextComment(comment)
  }
  /**
   * 
   * @param {string} html 
   * @param {string} text 
   * @param {CGSize} size 
   * @param {string} tag 
   */
  appendHtmlComment(html, text, size, tag){
    this.note.appendHtmlComment(html, text, size, tag)
  }
  /**
   * 
   * @param  {string[]} comments 
   * @returns 
   */
  appendMarkdownComments(...comments) {
    comments = unique(comments, true)
    const existComments = this.note.comments.filter(k => k.type === "TextNote")
    comments.forEach(comment => {
      if (
        comment &&
        existComments.every(k => k.type === "TextNote" && k.text !== comment)
      ) {
        if (this.note.appendMarkdownComment)
          this.note.appendMarkdownComment(comment)
        else this.note.appendTextComment(comment)
      }
    })
    return this
  }
  /**
   * 
   * @param {number} index 
   */
  removeCommentByIndex(index){
    this.note.removeCommentByIndex(index)
  }
  /**
   * Remove all comment but tag, link and also the filterd. And tags and links will be sat at the end。
   * @param filter not deleted
   * @param f call a function after deleted, before set tag and link
   */
  async removeCommentButLinkTag(
    // 不删除
    filter,
    f
  ) {
    const { removedIndex, linkTags } = this.note.comments.reduce(
      (acc, comment, i) => {
        if (
          comment.type == "TextNote" &&
          (comment.text.includes("marginnote3app://note/") ||
            comment.text.startsWith("#"))
        ) {
          acc.linkTags.push(comment.text)
          acc.removedIndex.unshift(i)
        } else if (!filter(comment)) acc.removedIndex.unshift(i)
        return acc
      },
      {
        removedIndex: [],
        linkTags: []
      }
    )
    removedIndex.forEach(k => {
      this.note.removeCommentByIndex(k)
    })
    f && (await f(this))
    this.appendTextComments(...linkTags)
    return this
  }
  /**
   * @param {string[]} titles
   * append titles as much as you want
   */
  appendTitles(...titles) {
    const newTitle = MNUtil.unique([...this.titles, ...titles], true).join("; ")
    if (this.note.excerptText === this.note.noteTitle) {
      this.note.noteTitle = newTitle
      this.note.excerptText = newTitle
    } else {
      this.note.noteTitle = newTitle
    }
    return this
  }
  /**
   * @param {string[]} tags
   * append tags as much as you want
   */
  appendTags(...tags) {
    this.tidyupTags()
    tags = unique([...this.tags, ...tags], true)
    const lastComment = this.note.comments[this.note.comments.length - 1]
    if (lastComment?.type == "TextNote" && lastComment.text.startsWith("#")) {
      this.note.removeCommentByIndex(this.note.comments.length - 1)
    }
    this.appendTextComments(tags.map(k => `#${k}`).join(" "))
    return this
  }
  /**
   * make sure tags are in the last comment
   */
  tidyupTags() {
    const existingTags= []
    const tagCommentIndex = []
    this.note.comments.forEach((comment, index) => {
      if (comment.type == "TextNote" && comment.text.startsWith("#")) {
        const tags = comment.text.split(" ").filter(k => k.startsWith("#"))
        existingTags.push(...tags.map(tag => tag.slice(1)))
        tagCommentIndex.unshift(index)
      }
    })

    tagCommentIndex.forEach(index => {
      this.note.removeCommentByIndex(index)
    })

    this.appendTextComments(
      MNUtil.unique(existingTags)
        .map(k => `#${k}`)
        .join(" ")
    )
    return this
  }
  /**
   * @param {MbBookNote | string} comment
   * get comment index by comment
   */
  getCommentIndex(comment) {
    const comments = this.note.comments
    for (let i = 0; i < comments.length; i++) {
      const _comment = comments[i]
      if (typeof comment == "string") {
        if (_comment.type == "TextNote" && _comment.text == comment) return i
      } else if (
        _comment.type == "LinkNote" &&
        _comment.noteid == comment.noteId
      )
        return i
    }
    return -1
  }
  clearFormat(){
    this.note.clearFormat()
  }
  /**
   * 
   * @param {MNNote|MbBookNote} note 
   * @param {string} type 
   */
  appendNoteLink(note,type="To"){
    switch (MNUtil.typeOf(note)) {
      case "MNNote":
        switch (type) {
          case "Both":
            this.note.appendNoteLink(note.note)
            note.note.appendNoteLink(this.note)
            break;
          case "To":
            this.note.appendNoteLink(note.note)
            break;
          case "From":
            note.note.appendNoteLink(this.note)
          default:
            break;
        }
        break;
      case "MbBookNote":
        switch (type) {
          case "Both":
            this.note.appendNoteLink(note)
            note.appendNoteLink(this.note)
            break;
          case "To":
            this.note.appendNoteLink(note)
            break;
          case "From":
            note.appendNoteLink(this.note)
          default:
            break;
        }
        break;
      default:
        break;
    }
  }
  clone() {
    let noteIds = MNUtil.db.cloneNotesToTopic([this.note], this.note.notebookId)
    return MNNote.new(noteIds[0])
  }
  focusInMindMap(){
    MNUtil.studyController.focusNoteInMindMapById(this.noteId)
  }
  focusInDocument(){
    MNUtil.studyController.focusNoteInDocumentById(this.noteId)
  }
  focusInFloatMindMap(){
    MNUtil.studyController.focusNoteInFloatMindMapById(this.noteId)
  }
  /**
   * 
   * @param {MbBookNote} note 
   * @returns 
   */
  static getNoteExcerptTextPic(note) {
    const acc = {
      ocr: [],
      html: [],
      md: []
    }
    const text = note.excerptText?.trim()
    if (note.excerptPic) {
      const imgs = MNNote.exportPic(note.excerptPic)
      if (imgs)
        Object.entries(imgs).forEach(([k, v]) => {
          if (k in acc) acc[k].push(v)
        })
      if (text) {
        acc.ocr.push(text)
      }
    } else {
      if (text) {
        Object.values(acc).forEach(k => k.push(text))
      }
    }
    return acc
  }
  /**
   * Get picture base64 code wrapped in html and markdown
   * @param {MNPic} pic
   * @returns
   * - html: `<img class="MNPic" src="data:image/jpeg;base64,${base64}"/>`
   * - md: `![MNPic](data:image/jpeg;base64,${base64})`
   */
  static exportPic(pic) {
    const base64 = MNUtil.db.getMediaByHash(pic.paint)?.base64Encoding()
    if (base64)
      return {
        html: `<img class="MNPic" src="data:image/jpeg;base64,${base64}"/>`,
        md: `![MNPic](data:image/jpeg;base64,${base64})`
      }
  }
  static focusInMindMapById(noteId){
    MNUtil.focusNoteInMindMapById(noteId)
  }
  static focusInDocumentById(noteId){
    MNUtil.focusNoteInDocumentById(noteId)
  }
  /**
   * 
   * @param {MbBookNote|string} note 
   */
  static focusInMindMap(note){
    switch (MNUtil.typeOf(note)) {
      case "MbBookNote":
        MNUtil.focusNoteInMindMapById(note.noteId)
        break;
      case "NoteURL":
        MNUtil.focusNoteInMindMapById(MNUtil.getNoteIdByURL(note))
        break;
      case 'string':
        MNUtil.focusNoteInMindMapById(note)
        break;
      default:
        break;
    }
  }
  /**
   * 
   * @param {MbBookNote|string} note 
   */
  static focusInDocument(note){
    switch (MNUtil.typeOf(note)) {
      case "MbBookNote":
        MNUtil.focusNoteInDocumentById(note.noteId)
        break;
      case "NoteURL":
        MNUtil.focusNoteInDocumentById(MNUtil.getNoteIdByURL(note))
        break;
      case 'string':
        MNUtil.focusNoteInDocumentById(note)
        break;
      default:
        break;
    }
  }
  static focusInFloatMindMap(note){
    switch (MNUtil.typeOf(note)) {
      case "MbBookNote":
        MNUtil.focusNoteInFloatMindMapById(note.noteId)
        break;
      case "NoteURL":
        MNUtil.focusNoteInFloatMindMapById(MNUtil.getNoteIdByURL(note))
        break;
      case 'string':
        MNUtil.focusNoteInFloatMindMapById(note)
        break;
      default:
        break;
    }
  }
  static getFocusNote() {
    let notebookController = MNUtil.notebookController
    if (!notebookController.view.hidden && notebookController.mindmapView && notebookController.focusNote) {
      return MNNote.new(notebookController.focusNote)
    }else{
      return MNNote.new(MNUtil.currentDocController.focusNote)
    }
  }

  static getFocusNotes() {
    let notebookController = MNUtil.notebookController
    if (!notebookController.view.hidden && notebookController.mindmapView && notebookController.mindmapView.selViewLst.length) {
      let selViewLst = notebookController.mindmapView.selViewLst
      return selViewLst.map(tem=>{
        return MNNote.new(tem.note.note)
      })
    }else if(MNUtil.currentDocController.focusNote){
      return [MNNote.new(MNUtil.currentDocController.focusNote)]
    }else{
      //如果两个上面两个都没有，那就可能是小窗里打开的
      return [MNNote.new(notebookController.focusNote)]
    }
  }
  static getSelectedNotes(){
    return this.getFocusNotes()
  }
  /**
   * 
   * @param {MbBookNote|MNNote|string} note 
   * @returns 
   */
  static clone(note) {

    let noteIds = []
    switch (MNUtil.typeOf(note)) {
      case "NoteURL":
        let noteFromURL = MNUtil.getNoteById(MNUtil.getNoteIdByURL(note))
        if (!noteFromURL) {
          MNUtil.showHUD("Note not exists!")
          return undefined
        }
        noteIds = MNUtil.db.cloneNotesToTopic([noteFromURL], noteFromURL.notebookId)
        return MNNote.new(noteIds[0])
      case "string":
        let targetNote = MNUtil.getNoteById(note)
        if (!targetNote) {
          MNUtil.showHUD("Note not exists!")
          return undefined
        }
        noteIds = MNUtil.db.cloneNotesToTopic([targetNote], targetNote.notebookId)
        return MNNote.new(noteIds[0])
      case "MbBookNote":
        noteIds = MNUtil.db.cloneNotesToTopic([note], note.notebookId)
        return MNNote.new(noteIds[0])
      case "MNNote":
        noteIds = MNUtil.db.cloneNotesToTopic([note.note], note.notebookId)
        return MNNote.new(noteIds[0])
      default:
        break;
    }
  }
}